# `envenom` - an elegant application configurator for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections.abc import Callable, Sequence
from dataclasses import dataclass, field
from typing import Any, TypeVar

from envenom.entries import (
    Entry,
    EntryWithDefault,
    EntryWithDefaultFactory,
    OptionalEntry,
    RequiredEntry,
)
from envenom.vars import Var

# Why so much 'type: ignore' around here?
#
# It's a hard truth to accept, but dataclasses are lying to us.
#
# When the `field` function is called it creates a Field[T] object, but it tells
# us that it really returns a T (or sometimes None), I swear, pretty please.
#
# Because we're exposing a similar interface and hooking into this exact layer,
# we therefore need to:
#   1) ignore the return value from `field`; it's a lie anyway
#   2) lie to our consumers on the API side just like `field` would, except of
#      course we supply three versions of the `field` function, so we need
#      to lie on all three fronts.


T = TypeVar("T")


def config(
    namespace: Sequence[str] | str | None = None,
) -> Callable[[type[T]], type[T]]:
    """
    Defines a new config class.

    Parameters:
        namespace:  Namespace to pull config values from. This allows different
                    configuration items with the same base name to exist in different
                    contexts. Can be specified either as a single `str` or a Sequence
                    of `str`s.

    Returns:
        A callable which transforms a class into a new config class.

    Example:
        Use this as a class decorator:

        ```python
        from envenom import config, required


        @config()
        class AppCfg:
            secret_key: str = required()
        ```

    """

    def __wrapper(cls: type[T]) -> type[T]:
        var: Var[Any]
        new_fields: dict[str, Any] = {
            name: field(
                init=False,
                repr=True,
                hash=True,
                default_factory=(var := entry.get_var(name, namespace)).get,
                metadata={
                    "config": cls,
                    "type": cls.__annotations__[name],
                    "var": var,
                },
            )
            for name, entry in cls.__dict__.items()
            if isinstance(entry, Entry) and name in cls.__annotations__
        }

        for name, f in new_fields.items():
            setattr(cls, name, f)

        return dataclass(frozen=True, eq=True)(cls)

    return __wrapper


def required(parser: Callable[[str], T] = str, *, file: bool = True) -> T:
    """
    Creates a required config field.

    Fields created with this function will have a type of `T`.

    Parameters:
        parser: Callable which will convert a single `str` into the
                desired object type.
        file:   Whether to try and retrieve the configuration value from a file
                using the environment variable with the `__FILE` suffix.

    Returns:
        An object of type `T` instantiated from the environment.

    Example:
        ```python
        from envenom import config, required


        @config()
        class AppCfg:
            secret_key: str = required()
        ```

    """

    return RequiredEntry(parser, file=file)  # type: ignore (2)


def optional(parser: Callable[[str], T] = str, *, file: bool = True) -> T | None:
    """
    Creates an optional config field.

    Fields created with this function will have a type of `T | None`.

    Parameters:
        parser: Callable which will convert a single `str` into the
                desired object type.
        file:   Whether to try and retrieve the configuration value from a file
                using the environment variable with the `__FILE` suffix.

    Returns:
        An object of type `T` instantiated from the environment if available,
            otherwise `None`.

    Example:
        ```python
        from envenom import config, optional


        @config()
        class AppCfg:
            signature: str | None = optional()
        ```

    """

    return OptionalEntry(parser, file=file)  # type: ignore (2)


def with_default(
    parser: Callable[[str], T] = str, *, file: bool = True, default: T
) -> T:
    """
    Creates an optional config field with a default value.

    Fields created with this function will have a type of `T`.

    Parameters:
        parser:     Callable which will convert a single `str` into the
                    desired object type.
        file:       Whether to try and retrieve the configuration value from a file
                    using the environment variable with the `__FILE` suffix.
        default:    The default to return if value is not set in the environment.

    Returns:
        An object of type `T` instantiated from the environment if available,
            otherwise the default value.

    Example:
        ```python
        from envenom import config, with_default
        from envenom.parsers import bool_parser


        @config()
        class AppCfg:
            feature_flag: bool = with_default(bool_parser(), default=False)
        ```

    """

    return EntryWithDefault(parser, file=file, default=default)  # type: ignore (2)


def with_default_factory(
    parser: Callable[[str], T] = str,
    *,
    file: bool = True,
    default_factory: Callable[[], T],
) -> T:
    """
    Creates an optional config field with a default value factory.

    Fields created with this function will have a type of `T`.

    Parameters:
        parser:             Callable which will convert a single `str` into the
                            desired object type.
        file:               Whether to try and retrieve the configuration value from a
                            file using the environment variable with the `__FILE`
                            suffix.
        default_factory:    The default factory to call if value is not set in the
                            environment.

    Returns:
        An object of type `T` instantiated from the environment if available,
            otherwise the value created by calling the default factory.

    Example:
        ```python
        from uuid import UUID, uuid4

        from envenom import config, with_default_factory


        @config()
        class AppCfg:
            uuid: UUID = with_default_factory(UUID, default_factory=uuid4)
        ```

    """

    return EntryWithDefaultFactory(
        parser,
        file=file,
        default_factory=default_factory,
    )  # type: ignore (2)


def subconfig(cls: Callable[[], T], /) -> T:
    """
    Includes another config class as a field of the current config class.

    This is really a convenvience function for expressing a dataclass
    field with a default factory, but it expresses the intention better too.

    `subconfig(cls)` is equivalent to `field(default_factory=cls)`.

    This function, unlike others in this module, is compatible with the
    standard `@dataclass` decorator.

    Parameters:
        cls:  Subconfig class to include as a field in this config class.

    Returns:
        An instantiated object of the subconfig class.

    Example:
        ```python
        from envenom import config, required, subconfig


        @config()
        class SubCfg:
            secret_key: str = required()


        @config()
        class AppCfg:
            subcfg: SubCfg = subconfig(SubCfg)
        ```
    """

    return field(default_factory=cls)



# `envenom` - an elegant application configurator for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import TypedDict

from flask import Flask

from envenom import config, with_default
from envenom.parsers import bool_parser


@config()
class FlaskCfg:
    static_docs: bool = with_default(bool_parser(), default=True)
    interactive_docs: bool = with_default(bool_parser(), default=False)


@config(namespace="app")
class AppCfg:
    motd: str = with_default(default="This is the default message of the day.")


cfg = FlaskCfg()
app = Flask(__name__)


class IndexResponse(TypedDict):
    motd: str


@app.route("/")
def index() -> IndexResponse:
    return {"motd": AppCfg().motd}


if cfg.static_docs:

    @app.route("/redoc")
    def static_docs() -> str:
        return "Static docs would be here if Flask had them built-in."


if cfg.interactive_docs:

    @app.route("/docs")
    def interactive_docs() -> str:
        return "Interactive docs would be here if Flask had them built-in."



# `envenom` - an elegant application configurator for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass


class GenericError(Exception):
    """
    Base exception class for any errors raised by `envenom`.
    """

    def __str__(self) -> str:
        return "error"

    def __format__(self, __format_spec: str) -> str:
        match __format_spec:
            case "class":
                return self.__class__.__name__
            case _:
                return str(self)


@dataclass
class ConfigurationError(GenericError):
    """
    Base exception class for any errors raised by `envenom`
    relating to a specific configuration field.
    """

    name: str

    def __str__(self) -> str:
        return f"error for configuration '{self.name}'"

    def __format__(self, __format_spec: str) -> str:
        if __format_spec == "var":
            return self.name
        return super().__format__(__format_spec)


class MissingConfiguration(ConfigurationError):
    """
    Raised when a required configuration entry is missing in the environment.
    """

    def __str__(self) -> str:
        return f"missing value for configuration '{self.name}'"


@dataclass
class InvalidConfiguration(ConfigurationError):
    """
    Raised when the value for a specific field is invalid.

    Most commonly raised when `ValueError` is raised from a field parser.
    """

    value: str

    def __str__(self) -> str:
        return f"invalid value '{self.value}' for configuration '{self.name}'"

    def __format__(self, __format_spec: str) -> str:
        if __format_spec == "value":
            return self.value
        return super().__format__(__format_spec)


class ConfigurationFileUnreadable(InvalidConfiguration):
    """
    Raised when an environment variable with a `__FILE` suffix is set and
    the value cannot be read from the file.
    """

    def __str__(self) -> str:
        return f"unreadable file '{self.value}' for configuration '{self.name}'"



# `envenom` - an elegant application configurator for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, Generic, TypeVar

from envenom import config

T = TypeVar("T")


class Singleton(type, Generic[T]):
    __instances: dict[type[Any], Any] = {}

    def __call__(cls, *args: Any, **kwargs: Any) -> T:
        if cls not in cls.__instances:
            cls.__instances[cls] = super().__call__(*args, **kwargs)
        return cls.__instances[cls]


@config()
class DefaultMainCfg:
    pass


@config()
class CachedMainCfg(metaclass=Singleton["CachedMainCfg"]):
    pass


if __name__ == "__main__":
    default_cfg1 = DefaultMainCfg()
    default_cfg2 = DefaultMainCfg()
    cached_cfg1 = CachedMainCfg()
    cached_cfg2 = CachedMainCfg()

    print("id(default_cfg1) should not be same as id(default_cfg2)")
    print(f"{id(default_cfg1)=}")
    print(f"{id(default_cfg2)=}")
    print("id(cached_cfg1) should be same as id(cached_cfg2)")
    print(f"{id(cached_cfg1)=}")
    print(f"{id(cached_cfg2)=}")



# `envenom` - an elegant application configurator for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass, field
from functools import cached_property

from envenom import config, optional, required, subconfig


def derive_public_key(private_key: str) -> str:
    return private_key[:8]


@config("jwt")
class JWTCfg:
    issuer: str = field(default="https://example.com")
    private_key: str = required()
    public_key: str | None = optional()

    @cached_property
    def current_public_key(self) -> str:
        return self.public_key or derive_public_key(self.private_key)


@dataclass
class OAuth2Cfg:
    jwt: JWTCfg = subconfig(JWTCfg)  # subconfig can be used in @dataclass!
    supports_oidc: bool = field(default=True)


@config()
class ApplicationCfg:
    secret_key: str = required()
    oauth2: OAuth2Cfg = subconfig(OAuth2Cfg)


if __name__ == "__main__":
    cfg = ApplicationCfg()

    # fmt: off
    # flake8: noqa
    print(f"cfg.secret_key ({type(cfg.secret_key)}): {repr(cfg.secret_key)}")
    print(f"cfg.oauth2.supports_oidc ({type(cfg.oauth2.supports_oidc)}): {repr(cfg.oauth2.supports_oidc)}")
    print(f"cfg.oauth2.jwt.issuer ({type(cfg.oauth2.jwt.issuer)}): {repr(cfg.oauth2.jwt.issuer)}")
    print(f"cfg.oauth2.jwt.private_key ({type(cfg.oauth2.jwt.private_key)}): {repr(cfg.oauth2.jwt.private_key)}")
    print(f"cfg.oauth2.jwt.public_key ({type(cfg.oauth2.jwt.public_key)}): {repr(cfg.oauth2.jwt.public_key)}")
    print(f"cfg.oauth2.jwt.current_public_key ({type(cfg.oauth2.jwt.current_public_key)}): {repr(cfg.oauth2.jwt.current_public_key)}")
    # fmt: on



# `envenom` - an elegant application configurator for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass
from typing import Generic, TypeVar

from envenom import config, optional, required
from envenom.parsers import bool_parser, list_parser

T = TypeVar("T")


@dataclass
class Box(Generic[T]):
    item: T


def boxed_int(v: str) -> Box[int]:
    return Box(item=int(v))


def boxed_bytes(v: str) -> Box[bytes]:
    return Box(item=v.encode())


@config()
class MainCfg:
    boxed_int: Box[int] = required(boxed_int)
    boxed_bytes: Box[bytes] = required(boxed_bytes)

    default_boolean: bool = required(bool_parser())
    custom_boolean: bool = required(
        bool_parser(true_values={"mhm"}, false_values={"uhhuh"})
    )

    required_list: list[str] = required(list_parser())
    required_empty_list: list[str] = required(list_parser())
    optional_list: list[str] | None = optional(list_parser())
    optional_empty_list: list[str] | None = optional(list_parser())
    optional_not_provided_list: list[str] | None = optional(list_parser())
    custom_list: list[int] = required(list_parser(int, separator=";"))


if __name__ == "__main__":
    cfg = MainCfg()

    # fmt: off
    # flake8: noqa
    print(f"cfg.boxed_int ({type(cfg.boxed_int)}): {repr(cfg.boxed_int)}")
    print(f"cfg.boxed_int.item ({type(cfg.boxed_int.item)}): {repr(cfg.boxed_int.item)}")
    print(f"cfg.boxed_bytes ({type(cfg.boxed_bytes)}): {repr(cfg.boxed_bytes)}")
    print(f"cfg.boxed_bytes.item ({type(cfg.boxed_bytes.item)}): {repr(cfg.boxed_bytes.item)}")
    print(f"cfg.default_boolean ({type(cfg.default_boolean)}): {repr(cfg.default_boolean)}")
    print(f"cfg.custom_boolean ({type(cfg.custom_boolean)}): {repr(cfg.custom_boolean)}")
    print(f"cfg.required_list ({type(cfg.required_list)}): {repr(cfg.required_list)}")
    print(f"cfg.required_empty_list ({type(cfg.required_empty_list)}): {repr(cfg.required_empty_list)}")
    print(f"cfg.optional_list ({type(cfg.optional_list)}): {repr(cfg.optional_list)}")
    print(f"cfg.optional_empty_list ({type(cfg.optional_empty_list)}): {repr(cfg.optional_empty_list)}")
    print(f"cfg.optional_not_provided_list ({type(cfg.optional_not_provided_list)}): {repr(cfg.optional_not_provided_list)}")
    print(f"cfg.custom_list ({type(cfg.custom_list)}): {repr(cfg.custom_list)}")
    # fmt: on



# `envenom` - an elegant application configurator for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from uuid import UUID, uuid4

from envenom import config, optional, required, with_default_factory


@config()
class UUIDCfg:
    admin_user_uuid: UUID = required(UUID)
    manager_user_uuid: UUID | None = optional(UUID)
    random_uuid_if_unset: UUID = with_default_factory(UUID, default_factory=uuid4)


if __name__ == "__main__":
    cfg = UUIDCfg()

    # fmt: off
    # flake8: noqa
    print(f"cfg.admin_user_uuid ({type(cfg.admin_user_uuid)}): {repr(cfg.admin_user_uuid)}")
    print(f"cfg.manager_user_uuid ({type(cfg.manager_user_uuid)}): {repr(cfg.manager_user_uuid)}")
    print(f"cfg.random_uuid_if_unset ({type(cfg.random_uuid_if_unset)}): {repr(cfg.random_uuid_if_unset)}")
    # fmt: on



# `envenom` - an elegant application configurator for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from envenom import config, required, subconfig


@config(("app", "cfg"))
class OtherNamespaceCfg:
    some_value: str = required()


@config(("ns1", "ns-2"))
class NestedNamespaceCfg:
    some_value: str = required()


@config("ns1")
class SimpleNamespaceCfg:
    some_value: str = required()
    nested_namespace: NestedNamespaceCfg = subconfig(NestedNamespaceCfg)


@config()
class MainCfg:
    some_value: str = required()
    simple_namespace: SimpleNamespaceCfg = subconfig(SimpleNamespaceCfg)
    other_namespace: OtherNamespaceCfg = subconfig(OtherNamespaceCfg)


if __name__ == "__main__":
    cfg = MainCfg()

    # fmt: off
    # flake8: noqa
    print(f"cfg.some_value ({type(cfg.some_value)}): {repr(cfg.some_value)}")
    print(f"cfg.simple_namespace.some_value ({type(cfg.simple_namespace.some_value)}): {repr(cfg.simple_namespace.some_value)}")
    print(f"cfg.simple_namespace.nested_namespace.some_value ({type(cfg.simple_namespace.nested_namespace.some_value)}): {repr(cfg.simple_namespace.nested_namespace.some_value)}")
    print(f"cfg.other_namespace.some_value ({type(cfg.other_namespace.some_value)}): {repr(cfg.other_namespace.some_value)}")
    # fmt: on



# `envenom` - an elegant application configurator for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import enum

from envenom import config, optional, required, with_default


class ExitCode(enum.IntEnum):
    OK = 0
    MISSING_CONFIG = 1
    INVALID_CONFIG = 2
    CONFIG_FILE_UNREADABLE = 3


class LaunchCode(enum.StrEnum):
    OK = enum.auto()
    LAUNCHPAD_OBSTRUCTED = enum.auto()
    NOT_ENOUGH_FUEL = enum.auto()
    OVERRIDDEN = enum.auto()


class DaVinciCode(enum.Enum):
    UNSOLVED = enum.auto()
    SOLVED = enum.auto()


@config()
class EnumCfg:
    exit_code: ExitCode = required(lambda c: ExitCode(int(c)))
    launch_code: LaunchCode | None = optional(lambda c: LaunchCode(c))
    davinci_code: DaVinciCode = with_default(
        lambda c: DaVinciCode(int(c)), default=DaVinciCode.UNSOLVED
    )


if __name__ == "__main__":
    cfg = EnumCfg()

    print(f"cfg.exit_code ({type(cfg.exit_code)}): {repr(cfg.exit_code)}")
    print(f"cfg.launch_code ({type(cfg.launch_code)}): {repr(cfg.launch_code)}")
    print(f"cfg.davinci_code ({type(cfg.davinci_code)}): {repr(cfg.davinci_code)}")



# `envenom` - an elegant application configurator for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from functools import cached_property
from uuid import UUID, uuid4

from envenom import (
    config,
    optional,
    required,
    subconfig,
    with_default,
    with_default_factory,
)
from envenom.parsers import bool_parser


@config(namespace=("myapp", "db"))
class DbCfg:
    scheme: str = with_default(default="postgresql+psycopg://")
    host: str = required()
    port: int = with_default(int, default=5432)
    database: str = required()
    username: str | None = optional()
    password: str | None = optional()
    connection_timeout: int | None = optional(int)
    sslmode_require: bool = with_default(bool_parser(), default=False)

    @cached_property
    def auth(self) -> str:
        if not self.username and not self.password:
            return ""

        auth = ""
        if self.username:
            auth += self.username
        if self.password:
            auth += f":{self.password}"
        if auth:
            auth += "@"

        return auth

    @cached_property
    def query_string(self) -> str:
        query: dict[str, str] = {}
        if self.connection_timeout:
            query["timeout"] = str(self.connection_timeout)
        if self.sslmode_require:
            query["sslmode"] = "require"

        if not query:
            return ""

        query_string = "&".join((f"{key}={value}" for key, value in query.items()))
        return f"?{query_string}"

    @cached_property
    def connection_string(self) -> str:
        return (
            f"{self.scheme}{self.auth}{self.host}:{self.port}"
            f"/{self.database}{self.query_string}"
        )


@config(namespace="myapp")
class AppCfg:
    worker_id: UUID = with_default_factory(UUID, default_factory=uuid4)
    secret_key: str = required()
    db: DbCfg = subconfig(DbCfg)


if __name__ == "__main__":
    cfg = AppCfg()

    # fmt: off
    # flake8: noqa
    print(f"cfg.worker_id ({type(cfg.worker_id)}): {repr(cfg.worker_id)}")
    print(f"cfg.secret_key ({type(cfg.secret_key)}): {repr(cfg.secret_key)}")
    print(f"cfg.db.host ({type(cfg.db.host)}): {repr(cfg.db.host)}")
    print(f"cfg.db.port ({type(cfg.db.port)}): {repr(cfg.db.port)}")
    print(f"cfg.db.database ({type(cfg.db.database)}): {repr(cfg.db.database)}")
    print(f"cfg.db.username ({type(cfg.db.username)}): {repr(cfg.db.username)}")
    print(f"cfg.db.password ({type(cfg.db.password)}): {repr(cfg.db.password)}")
    print(f"cfg.db.connection_timeout ({type(cfg.db.connection_timeout)}): {repr(cfg.db.connection_timeout)}")
    print(f"cfg.db.sslmode_require ({type(cfg.db.sslmode_require)}): {repr(cfg.db.sslmode_require)}")
    print(f"cfg.db.connection_string ({type(cfg.db.connection_string)}): {repr(cfg.db.connection_string)}")
    # fmt: on



# `envenom` - an elegant application configurator for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections.abc import Callable, Collection
from typing import TypeVar

T = TypeVar("T")


def bool_parser(
    true_values: Collection[str] = {"t", "true", "y", "yes", "1", "+", "✓"},
    false_values: Collection[str] = {"f", "false", "n", "no", "0", "-", "✗"},
) -> Callable[[str], bool]:
    """
    This handles one of the most basic conversions: `str` -> `bool`.

    `bool`, unlike e.g. `int`, will not just do the right thing,
    so we need a specialized parser.

    This parser is case-insensitive.

    Parameters:
        true_values:    Collection of values which will evaluate to `True`.
        false_values:   Collection of values which will evaluate to `False`.

    Returns:
        A parser callable which will parse the `str` into a `bool`.

    Example:
        ```python
        from envenom import config, with_default
        from envenom.parsers import bool_parser


        @config()
        class AppCfg:
            feature_flag: bool = with_default(bool_parser(), default=False)
        ```

    """

    ts = set(map(lambda s: s.lower(), true_values))
    fs = set(map(lambda s: s.lower(), false_values))

    def __parser(v: str) -> bool:
        normalized = v.lower()
        if normalized in ts:
            return True
        if normalized in fs:
            return False
        raise ValueError(v)

    return __parser


def list_parser(
    parser: Callable[[str], T] = str, *, separator: str = ","
) -> Callable[[str], list[T]]:
    """
    This handles one of the most basic conversions: lists of items.

    Parameters:
        parser:     Parser applied to each list element.
        separator:  Separator for splitting the list.

    Returns:
        A parser callable which will parse the `str` into a list of objects of
            type `T` resulting from splitting the value using the delimiter
            and applying the parser function on each element.

    Example:
        ```python
        from uuid import UUID

        from envenom import config, required
        from envenom.parsers import list_parser


        @config()
        class AppCfg:
            admin_uuids: list[UUID] = required(list_parser(UUID, separator=";"))
        ```

    """

    def __parser(v: str) -> list[T]:
        if not v:
            return []
        return [parser(s) for s in v.split(separator)]

    return __parser



# `envenom` - an elegant application configurator for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass

from fastapi import FastAPI
from fastapi.routing import APIRoute

from envenom import config, with_default
from envenom.parsers import bool_parser


@config()
class FastAPICfg:
    static_docs: bool = with_default(bool_parser(), default=True)
    interactive_docs: bool = with_default(bool_parser(), default=False)


@config(namespace="app")
class AppCfg:
    motd: str = with_default(default="This is the default message of the day.")


@dataclass
class IndexResponse:
    motd: str


async def get_index() -> IndexResponse:
    return IndexResponse(motd=AppCfg().motd)


cfg = FastAPICfg()
app = FastAPI(
    title="envenom.examples.fastapi",
    summary="An elegant application configurator for the more civilized age",
    redoc_url="/redoc" if cfg.static_docs else None,
    docs_url="/docs" if cfg.interactive_docs else None,
    routes=[
        APIRoute("/", get_index, methods={"GET"}),
    ],
)



# `envenom` - an elegant application configurator for the more civilized age
# Copyright (C) 2024 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import re
from abc import ABCMeta
from collections.abc import Callable, Sequence
from dataclasses import dataclass, field
from functools import cached_property
from typing import ClassVar, Generic, TypeAlias, TypeVar

from envenom.errors import (
    ConfigurationFileUnreadable,
    InvalidConfiguration,
    MissingConfiguration,
)

T = TypeVar("T")

Namespace: TypeAlias = Sequence[str] | str | None

Parser: TypeAlias = Callable[[str], T]


def get_env_var_value(name: str) -> str | None:
    return os.environ.get(name)


def get_file_var_value(name: str) -> str | None:
    with open(name, "r") as f:
        return f.read().strip() or None


@dataclass
class Var(Generic[T], metaclass=ABCMeta):
    name: str
    namespace: Namespace = field(default=None)
    parser: Parser[T] = field(repr=False, kw_only=True)
    file: bool = field(repr=False, kw_only=True, default=True)

    env_name_pattern: ClassVar[re.Pattern[str]] = re.compile("[^0-9a-zA-Z_]+")

    def get(self) -> T | None:
        if (v := self.get_raw_value()) is None:
            return None

        try:
            return self.parser(v)
        except ValueError as e:
            raise InvalidConfiguration(self.env_var_name, v) from e

    def get_raw_value(self) -> str | None:
        if (
            self.file
            and (file := get_env_var_value(self.file_env_var_name)) is not None
        ):
            try:
                return get_file_var_value(file)
            except IOError as e:
                raise ConfigurationFileUnreadable(self.env_var_name, file) from e

        return get_env_var_value(self.env_var_name)

    @cached_property
    def normalized_namespace(self) -> str | None:
        if self.namespace is None:
            return None

        if isinstance(self.namespace, str):
            namespace = (self.namespace,)
        else:
            namespace = self.namespace

        return "__".join(
            map(lambda s: re.sub(self.env_name_pattern, "_", s).upper(), namespace)
        )

    @cached_property
    def namespace_prefix(self) -> str:
        return (
            f"{self.normalized_namespace}__"
            if self.normalized_namespace is not None
            else ""
        )

    @cached_property
    def normalized_name(self) -> str:
        return re.sub(self.env_name_pattern, "_", self.name).upper()

    @cached_property
    def env_var_name(self) -> str:
        return f"{self.namespace_prefix}{self.normalized_name}"

    @cached_property
    def file_env_var_name(self) -> str:
        return f"{self.env_var_name}__FILE"


class RequiredVar(Var[T]):
    def get(self) -> T:
        if (value := super().get()) is None:
            raise MissingConfiguration(self.env_var_name)
        return value


class OptionalVar(Var[T]):
    pass


@dataclass
class VarWithDefault(Generic[T], Var[T]):
    default: T = field(repr=False, kw_only=True)

    def get(self) -> T:
        return value if (value := super().get()) is not None else self.default


@dataclass
class VarWithDefaultFactory(Generic[T], Var[T]):
    default_factory: Callable[[], T] = field(repr=False, kw_only=True)

    def get(self) -> T:
        return value if (value := super().get()) is not None else self.default_factory()
